"use server";

import { getServerSession } from "next-auth";
import { AliasEntry, approveAliasEntry, createAliasEntry, createUserEntry, database, deleteAliasEntry, getAlias, getAllAliases, getUserAliases, setUserPassword } from "./db";
import { aliasesNeedApproval, isAdmin } from "./util";
import { AuditLog, auditLog } from "./audit";
import fs from "fs/promises";

export async function fetchAllUsers(): Promise<string[]> {
  return new Promise(async (resolve, reject) => {
    const session = await getServerSession();
    if (!isAdmin(session)) return reject("Unauthenticated");

    const db = database('credentials');
    db.all("SELECT key FROM passwords", (err, res: any) => {
      if (err) return reject(err);

      resolve(res.map((row: any) => row.key));
    });
  });
}

export async function changeOwnPassword(newPass: string) {
  if (newPass.length < 8) throw new Error("Invalid password");
  const session = await getServerSession();
  if (!session?.user?.email) throw new Error("Unauthenticated");
  await setUserPassword(session.user.email, newPass);
  auditLog("changeOwnPassword");
}

export async function fetchOwnAliases() {
  const session = await getServerSession();
  if (!session?.user?.email) throw new Error("Unauthenticated");
  return await getUserAliases(session.user.email);
}

export async function fetchUserAliases(email: string) {
  const session = await getServerSession();
  if (!session?.user?.email) throw new Error("Unauthenticated");
  if (!isAdmin(session)) throw new Error("Unauthorized");

  return await getUserAliases(email);
}

export async function fetchAllAliases() {
  const session = await getServerSession();
  if (!session?.user?.email) throw new Error("Unauthenticated");
  if (!isAdmin(session)) throw new Error("Unauthorized");

  return await getAllAliases();
}

export async function aliasAvailable(email: string) {
  const session = await getServerSession();
  if (!session?.user) throw new Error("Unauthenticated");

  return new Promise<boolean>((resolve, reject) => {
    const db = database('aliases');
    db.get('SELECT id FROM aliases WHERE alias = ?', email.toLowerCase(), (err, res) => {
      db.close();
      if (err) return reject(err);
      if (res != undefined) return resolve(false);

      const authDb = database('credentials');
      authDb.get('SELECT key FROM passwords WHERE key = ?', email.toLowerCase(), (err, res) => {
        authDb.close();
        if (err) return reject(err);
        resolve(res == undefined);
      });
    });
  });
}

export async function createAlias(user: string, alias: string): Promise<AliasEntry> {
  const session = await getServerSession();
  if (!session?.user?.email) throw new Error("Unauthenticated");
  if (!isAdmin(session)) throw new Error("Unauthenticated");
  if (!await aliasAvailable(alias)) throw new Error("Alias unavailable");

  const id = await createAliasEntry(user, alias.toLowerCase(), false);
  const res = {
    id: id,
    address: session.user.email,
    alias: alias,
    pending: false,
  };

  auditLog('createAlias', res);
  return res;
}

export async function createAliasSelf(alias: string): Promise<AliasEntry> {
  const session = await getServerSession();
  if (!session?.user?.email) throw new Error("Unauthenticated");

  const pending = aliasesNeedApproval(session);
  if (!await aliasAvailable(alias)) throw new Error("Alias unavailable");

  const id = await createAliasEntry(session.user.email, alias.toLowerCase(), pending);
  const res = {
    id: id,
    address: session.user.email,
    alias: alias,
    pending: pending
  };

  auditLog('requestAlias', res);
  return res;
}

export async function deleteAlias(alias: string) {
  const session = await getServerSession();
  if (!session?.user?.email) throw new Error("Unauthenticated");

  if (!isAdmin(session) && (await getAlias(alias))?.address != session.user.email) {
    throw new Error("Unauthorized");
  }

  await deleteAliasEntry(alias);
  auditLog('deleteAlias', alias);
}

export async function approveAlias(alias: string) {
  const session = await getServerSession();
  if (!session?.user?.email) throw new Error("Unauthenticated");
  if (!isAdmin(session)) throw new Error("Unauthorized");

  await approveAliasEntry(alias);
  auditLog('approveAlias', alias);
}

export async function createUser(email: string, password: string) {
  const session = await getServerSession();
  if (!session?.user?.email) throw new Error("Unauthenticated");
  if (!isAdmin(session)) throw new Error("Unauthorized");

  if (!await aliasAvailable(email.toLowerCase())) {
    throw new Error("Alias unavailable");
  }

  if (password.length < 8 || !email.includes("@")) {
    throw new Error("Validation failed");
  }

  await createUserEntry(email, password);
  auditLog('createUser', email);
}

export async function fetchAuditLog(page: number): Promise<{ page: number, perPage: number, totalItems: number, items: (AuditLog & { index: number })[] }> {
  const itemsPerPage = 10;

  const session = await getServerSession();
  if (!session?.user?.email) throw new Error("Unauthenticated");
  if (!isAdmin(session)) throw new Error("Unauthorized");

  if (!process.env.AUDIT_FILE_PATH) return { page: page, perPage: itemsPerPage, totalItems: 0, items: [] };

  const lines = (await fs.readFile(process.env.AUDIT_FILE_PATH))
    .toString("utf8")
    .trim()
    .split("\n")
    .reverse();

  if (page  > Math.floor(lines.length / itemsPerPage)) {
    page = Math.ceil(lines.length / itemsPerPage) - 1;
  } else if (page < 0) {
    page = 0;
  }

  return {
    page: page,
    perPage: itemsPerPage,
    totalItems: lines.length,
    items: lines
      .slice(itemsPerPage * page, itemsPerPage * (page + 1))
      .map((item, i) => ({ ...JSON.parse(item), index: page * itemsPerPage + i })),
  };
}
